# -*- coding: utf-8 -*-
"""
One thing to note: in the NuSpec XML file, 'id' is the project name
which must be unique across the NuGet server and 'title' is the
human-friendly title of the package.
"""

import datetime as dt
import json

from sqlalchemy import Column, Integer, String, Text, Boolean
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import desc
from sqlalchemy import or_
from sqlalchemy import cast
from sqlalchemy.dialects import sqlite
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from pynuget import logger


Base = declarative_base()


class Package(Base):
    """
    """

    __tablename__ = "package"

    package_id = Column(Integer, primary_key=True)
    name = Column(String(256), index=True, unique=True, nullable=False)
    title = Column(String(256))
    download_count = Column(Integer, index=True, nullable=False, default=0)
    latest_version = Column(Text())

    def __repr__(self):
        return "<Package({}, {})>".format(self.package_id, self.name)


class Version(Base):
    """
    """

    __tablename__ = "version"

    version_id = Column(Integer, primary_key=True)
    package_id = Column(Integer, ForeignKey("package.package_id"),
                        nullable=False)
    title = Column(Text())
    description = Column(Text())
    created = Column(Integer)
    version = Column(String(32), index=True)
    package_hash = Column(Text())
    package_hash_algorithm = Column(Text())
    dependencies = Column(Text())
    package_size = Column(Integer)
    release_notes = Column(Text())
    version_download_count = Column(Integer, nullable=False, default=0)
    tags = Column(Text())
    license_url = Column(Text())
    project_url = Column(Text())
    icon_url = Column(Text())
    authors = Column(Text())
    owners = Column(Text())
    require_license_acceptance = Column(Boolean())
    copyright_ = Column(Text())
    is_prerelease = Column(Boolean())

    package = relationship("Package", backref="versions")

    def __repr__(self):
        return "<Version({}, {})>".format(self.package.name, self.version)

    @hybrid_property
    def thing(self):
        return "{}~~{}".format(self.package_id, self.version)

    @thing.expression
    def thing(cls):
        name = cast(cls.package_id, String)
        version = cast(cls.version, String)
        return name + "~~" + version


def count_packages(session):
    """
    Count the number of packages on the server.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`

    Returns
    -------
    int
    """
    logger.debug("db.count_packages()")
    return session.query(func.count(Package.package_id)).scalar()


def search_packages(session,
                    include_prerelease=False,
                    order_by=desc(Version.version_download_count),
                    filter_=None,
                    search_query=None):
    """
    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    include_prerelease : bool
    order_by : :class:`sqlalchemy.sql.operators.ColumnOperators`
    filder_ : str
        One of ('is_absolute_latest_version', 'is_latest_version').
    search_query : str
    """
    logger.debug("db.search_packages(...)")
    query = session.query(Version).join(Package)

    if search_query is not None:
        search_query = "%" + search_query + "%"
        query = query.filter(
            or_(Package.name.like(search_query),
                Package.title.like(search_query)
                )
        )

    if not include_prerelease:
        query = query.filter(Version.is_prerelease.isnot(True))

    known_filters = ('is_absolute_latest_version', 'IsLatestVersion')
    if filter_ is None:
        pass
    elif filter_ in known_filters:
        query = query.filter(Version.version == Package.latest_version)
    else:
        raise ValueError("Unknown filter '{}'".format(filter_))

    if order_by is not None:
        query = query.order_by(order_by)

    results = query.all()
    logger.debug("Found %d results." % len(results))

    return results


def package_updates(session, packages_dict, include_prerelease=False):
    """
    I *think* this returns a list of packages that need updating...

    Seems like the route for this fuction expects args from a url that look
    like:
        /updates?packageids='pkg1'|'pkg2'&versions='vers1'|'vers2'
    where "|" might be encoded as %7C

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    packages_dict : dict
        Dict of {package.name, version}.
    include_prerelease : bool
    """
    logger.debug("db.package_updates(...)")
    package_versions = ["{}~~{}".format(pkg, vers)
                        for pkg, vers
                        in packages_dict.items()]

    query = (session.query(Version)
             .filter(Version.version == Package.latest_version)
             .filter(Version.package_id.in_(packages_dict.keys()))
             .filter(~Version.thing.in_(package_versions))
             )

    return query.order_by(Version.package_id).all()


def find_by_pkg_name(session, package_name, version=None):
    """
    Find a package by name. If version is `None`, returns all versions.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    package_name : str
        The NuGet name of the package - the "id" tag in the NuSpec file.
    version : str
        The version of the package to download. If `None`, then return all
        versions.

    Returns
    -------
    results : list of :class:`Version`
    """
    logger.debug("db.find_by_pkg_name('%s', version='%s')" % (package_name,
                                                              version))
    query = (session.query(Version)
             .join(Package)
             .filter(Package.name == package_name)
             )

    stmt = query.statement.compile(dialect=sqlite.dialect(),
                                   compile_kwargs={"literal_binds": True})
    logger.debug(stmt)

    if version:
        query = query.filter(Version.version == version)
    query.order_by(desc(Version.version))

    results = query.all()
    logger.info("Found %d results." % len(results))
    logger.debug(results)

    return results


def find_pkg_by_id(session, package_id):
    query = (session.query(Package)
             .filter(Package.package_id == package_id)
             )

    # TODO: Error handling
    result = query.one()
    logger.debug(result)

    return result


def validate_id_and_version(session, package_name, version):
    """
    Not exactly sure what this is supposed to do, but I *think* it simply
    makes sure that the given pacakge_id and version exist... So that's
    what I've decided to make it do.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    package_name : str
        The NuGet name of the package - the "id" tag in the NuSpec file.
    version : str
    """
    logger.debug("db.validate_id_and_version(...)")
    query = (session.query(Version)
             .filter(Package.name == package_name)
             .filter(Version.version == version)
             )
    return session.query(query.exists()).scalar()


def increment_download_count(session, package_name, version):
    """
    Increment the download count for a given package version.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    package_name : str
        The NuGet name of the package - the "id" tag in the NuSpec file.
    version : str
    """
    msg = "db.increment_download_count(%s, %s)"
    logger.debug(msg % (package_name, version))
    obj = (session.query(Version)
           .filter(Package.name == package_name)
           .filter(Version.version == version)
           ).one()
    obj.version_download_count += 1
    obj.package.download_count += 1
    session.commit()


def insert_or_update_package(session, package_name, title, latest_version):
    """

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    package_name : str
        The NuGet name of the package - the "id" tag in the NuSpec file.
    title : str
    latest_version : str
    """
    logger.debug("db.insert_or_update_package(...)")
    sql = session.query(Package).filter(Package.name == package_name)
    obj = sql.one_or_none()
    if obj is None:
        pkg = Package(name=package_name, title=title,
                      latest_version=latest_version)
        session.add(pkg)
    else:
        sql.update({Package.name: package_name,
                    Package.title: title,
                    Package.latest_version: latest_version})
    session.commit()


def insert_version(session, **kwargs):
    """Insert a new version of an existing package."""
    logger.debug("db.insert_version(...)")
    kwargs['created'] = dt.datetime.utcnow()
    if 'dependencies' in kwargs:
        kwargs['dependencies'] = json.dumps(kwargs['dependencies'])
    if 'is_prerelease' not in kwargs:
        kwargs['is_prerelease'] = 0
    if 'require_license_acceptance' not in kwargs:
        kwargs['require_license_acceptance'] = 0

    version = Version(**kwargs)
    session.add(version)
    session.commit()
    logger.debug(version)


def delete_version(session, package_name, version):
    """

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.session.Session`
    package_name : str
        The NuGet name of the package - the "id" tag in the NuSpec file.
    version : str
    """
    msg = "db.delete_version({}, {})"
    logger.debug(msg.format(package_name, version))
    sql = (session.query(Version).join(Package)
           .filter(Package.name == package_name)
           .filter(Version.version == version)
           )
    package = sql.one().package
    logger.debug(package)

    session.delete(sql.one())
    session.commit()

    # update the Package.latest_version value, or delete the Package
    versions = (session.query(Version)
                .filter(Package.name == package_name)
                ).all()
    if len(versions) > 0:
        package.latest_version = max(v.version for v in versions)
    else:
        session.delete(package)
    session.commit()



# -*- coding: utf-8 -*-
"""
"""
import re
from pathlib import Path
from uuid import uuid4

# Third-Party
from flask import current_app
from flask import g
from flask import request
from flask import send_file
from flask import make_response
from flask import Blueprint
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from werkzeug.local import LocalProxy

from pynuget import db
from pynuget import core
from pynuget import logger
from pynuget.feedwriter import FeedWriter
from pynuget.core import et_to_str
from pynuget.core import ApiException


FEED_CONTENT_TYPE_HEADER = 'application/atom+xml; type=feed; charset=UTF-8'


pages = Blueprint('pages', __name__)


def get_db_session():
    session = getattr(g, 'session', None)
    if session is None:
        # TODO: Handle MySQL/PostgreSQL backends.
        db_name = Path(current_app.config['SERVER_PATH']) / Path(current_app.config['DB_NAME'])

        # TODO: Move this check so that it gets run on server start, not
        # just when a route that uses the session is called.
        if not db_name.exists():
            msg = "'{}' does not exist. Did you forget to run pynuget init?"
            msg = msg.format(str(db_name))
            logger.critical(msg)
            raise FileNotFoundError(msg)

        url = "sqlite:///{}".format(str(db_name))
        engine = create_engine(url, echo=False)
        Session = sessionmaker(bind=engine)
        session = g.session = Session()
    return session


session = LocalProxy(get_db_session)


@pages.route('/$metadata')
def meta():
    """
    The `nuget list` command calls this route.
    """
    logger.debug("route: /$metadata")
    resp = make_response(
        """<?xml version="1.0" encoding="utf-8"?>
        <service xml:base="{}" xmlns="http://www.w3.org/2007/app" xmlns:atom="http://www.w3.org/2005/Atom">
          <workspace>
            <atom:title type="text">Default</atom:title>
            <collection href="Packages">
              <atom:title type="text">Packages</atom:title>
            </collection>
          </workspace>
        </service>""".format(request.url_root)
    )
    resp.headers['Content-Type'] = 'application/atomsvc+xml; charset=utf-8'
    resp.headers['Cache-Control'] = 'no-cache'
    resp.headers['Pragma'] = 'no-cache'
    resp.headers['Expires'] = '-1'

    logger.debug("Returning resp: %s" % str(resp))
    return resp


@pages.route('/', methods=['GET'])
@pages.route('/index', methods=['GET'])
def index():
    """
    Used for web interface.
    """
    logger.debug("Route: /index")
    return "index page"


@pages.route('/api/v2/package/', methods=['PUT'])
def push():
    """
    Used by `nuget push`.
    """
    logger.debug("push()")
    logger.debug("  args: {}".format(request.args))
    logger.debug("  header: {}".format(request.headers))
    if not core.require_auth(request.headers):
        return "api_error: Missing or Invalid API key", 401

    logger.debug("Checking for uploaded file.")
    if 'package' not in request.files:
        logger.error("Package file was not uploaded.")
        return "error: File not uploaded", 409
    file = request.files['package']

    # Save the file to a temporary location
    file = core.save_file(file, "_temp", str(uuid4()))

    # Open the zip file that was sent and extract out the .nuspec file."
    try:
        nuspec = core.extract_nuspec(file)
    except Exception as err:
        logger.error("Exception: %s" % err)
        return "api_error: Zero or multiple nuspec files found", 400

    # The NuSpec XML file uses namespaces.
    # TODO: What if the namespace changes?
    ns = {'nuspec': 'http://schemas.microsoft.com/packaging/2012/06/nuspec.xsd'}

    # Make sure both the ID and the version are provided in the .nuspec file.
    try:
        metadata, pkg_name, version = core.parse_nuspec(nuspec, ns)
    except ApiException as err:
        return str(err), 400
    except Exception as err:
        logger.error(err)
        return str(err), 400

    valid_id = re.compile('^[A-Z0-9\.\~\+\_\-]+$', re.IGNORECASE)

    # Make sure that the ID and version are sane
    if not re.match(valid_id, pkg_name) or not re.match(valid_id, version):
        logger.error("Invalid ID or version.")
        return "api_error: Invlaid ID or Version", 400

    # and that we don't already have that ID+version in our database
    if db.validate_id_and_version(session, pkg_name, version):
        logger.error("Package %s version %s already exists" % (pkg_name, version))
        return "api_error: Package version already exists", 409

    # Hash the uploaded file and encode the hash in Base64. For some reason.
    try:
        # rename our file.
        # Check if the package's directory already exists. Create if needed.
        new = file.parent.parent / pkg_name / (version + ".nupkg")
        core.create_parent_dirs(new)
        logger.debug("Renaming %s to %s" % (str(file), new))
        file.rename(new)
        hash_, filesize = core.hash_and_encode_file(str(new))
    except Exception as err:
        logger.error("Exception: %s" % err)
        return "api_error: Unable to save file", 500

    try:
        dependencies = core.determine_dependencies(metadata, ns)
    except Exception as err:
        logger.error("Exception: %s" % err)
        return "api_error: Unable to parse dependencies.", 400

    logger.debug(dependencies)

    # and finaly, update our database.
    logger.debug("Updating database entries.")

    db.insert_or_update_package(session,
                                package_name=pkg_name,
                                title=et_to_str(metadata.find('nuspec:title', ns)),
                                latest_version=version)
    pkg_id = (session.query(db.Package)
              .filter(db.Package.name == pkg_name).one()
              ).package_id
    logger.debug("package_id = %d" % pkg_id)
    db.insert_version(
        session,
        authors=et_to_str(metadata.find('nuspec:authors', ns)),
        copyright_=et_to_str(metadata.find('nuspec:copyright', ns)),
        dependencies=dependencies,
        description=et_to_str(metadata.find('nuspec:description', ns)),
        package_hash=hash_,
        package_hash_algorithm='SHA512',
        package_size=filesize,
        icon_url=et_to_str(metadata.find('nuspec:iconUrl', ns)),
        is_prerelease='-' in version,
        license_url=et_to_str(metadata.find('nuspec:licenseUrl', ns)),
        owners=et_to_str(metadata.find('nuspec:owners', ns)),
        package_id=pkg_id,
        project_url=et_to_str(metadata.find('nuspec:projectUrl', ns)),
        release_notes=et_to_str(metadata.find('nuspec:releaseNotes', ns)),
        require_license_acceptance=et_to_str(metadata.find('nuspec:requireLicenseAcceptance', ns)) == 'true',
        tags=et_to_str(metadata.find('nuspec:tags', ns)),
        title=et_to_str(metadata.find('nuspec:id', ns)),
        version=version,
    )

    logger.info("Sucessfully updated database entries for package %s version %s." % (pkg_name, version))

    resp = make_response('', 201)
    return resp


@pages.route('/count', methods=['GET'])
def count():
    """
    Not sure which nuget command uses this...
    """
    logger.debug("Route: /count")
    resp = make_response(str(db.count_packages(session)))
    resp.headers['Content-Type'] = 'text/plain; charset=utf-8'
    return resp


@pages.route('/delete', methods=['DELETE'])
@pages.route('/<package>/<version>', methods=['DELETE'])
@pages.route('/api/v2/package/<package>/<version>', methods=['DELETE'])
def delete(package=None, version=None):
    """
    Used by `nuget delete`.
    """
    logger.debug("Route: /delete")
    if not core.require_auth(request.headers):
        return "api_error: Missing or Invalid API key", 401      # TODO

    if package is not None:
        pkg_name = package
    else:
        pkg_name = request.args.get('id')

    if version is None:
        version = request.args.get('version')
    path = core.get_package_path(pkg_name, version)
    path = Path(current_app.config['SERVER_PATH']) / current_app.config['PACKAGE_DIR'] / path

    if path.exists():
        path.unlink()

    try:
        db.delete_version(session, pkg_name, version)
    except NoResultFound:
        msg = "Version '{}' of Package '{}' was not found."
        return msg.format(pkg_name, version), 404

    logger.info("Sucessfully deleted package %s version %s." % (pkg_name, version))

    return '', 204


@pages.route('/download/<pkg_id>/<version>', methods=['GET'])
def download(pkg_id=None, version=None):
    """
    Indirectly used by `nuget install`.
    """
    logger.debug("Route: /download")

    if pkg_id is None:
        pkg_id = request.args.get('Id')
    if version is None:
        version = request.args.get('Version')

    pkg_name = db.find_pkg_by_id(session, pkg_id).name

    path = core.get_package_path(pkg_name, version)

    # Make sure we're trying to download from our package direcory
    path = Path(current_app.config['SERVER_PATH']) / current_app.config['PACKAGE_DIR'] / path
    abs_path = Path.cwd() / path

    logger.debug("Path to package: %s" % abs_path)
    db.increment_download_count(session, pkg_name, version)
    filename = "{}.{}.nupkg".format(pkg_name, version)
    logger.debug("File name: %s" % filename)

    logger.debug("sending file")
    result = send_file(str(abs_path),
                       mimetype="application/zip",
                       as_attachment=True,
                       attachment_filename=filename)

    header_str = str(result.headers).replace("\r\n", "\r\n  ").strip()
    logger.debug("Header: \n  {}".format(header_str))
    return result, 200


@pages.route('/FindPackagesById()', methods=['GET'])
@pages.route('/Packages(<func_args>)', methods=['GET'])
def find_by_id(func_args=None):
    """
    Used by `nuget install`.

    It looks like the NuGet client expects this to send back a list of all
    versions for the package, and then the client handles extraction of
    an individual version.

    Note that this is different from the newer API
        /Packages(Id='pkg_name',Version='0.1.3')
    which appears to move the version selection to server-side.
    """
    logger.debug("Route: /find_by_id")
    logger.debug("  args: {}".format(request.args))
    logger.debug("  header: {}".format(request.headers))

    if func_args is not None:
        # Using the newer API
        logger.debug(func_args)
        # Parse the args. From what I can tell, the only things sent here are:
        #   'Id': the name of the package
        #   'Version': the version string.
        # Looks like:
        #   "Id='NuGetTest',Version='0.0.2'"
        # Naive implementation
        regex = re.compile(r"^Id='(?P<name>.+)',Version='(?P<version>.+)'$")
        match = regex.search(func_args)
        if match is None:
            msg = "Unable to parse the arg string `{}`!"
            logger.error(msg.format(func_args))
            return msg.format(func_args), 500

        pkg_name = match.group('name')
        version = match.group('version')
        logger.debug("{}, {}".format(pkg_name, version))
    else:
        # old API
        pkg_name = request.args.get('id')
        sem_ver_level = request.args.get('semVerLevel', default=None)
        version = None

    # Some terms are quoted
    pkg_name = pkg_name.strip("'")

    results = db.find_by_pkg_name(session, pkg_name, version)
    logger.debug(results)
    feed = FeedWriter('FindPackagesById', request.url_root)
    resp = make_response(feed.write_to_output(results))
    resp.headers['Content-Type']

    logger.debug(resp.data.decode('utf-8').replace('><',' >\n<'))

    return resp


@pages.route('/Search()', methods=['GET'])
def search():
    """
    Used by `nuget list`.
    """
    logger.debug("Route: /search")
    logger.debug(request.args)
    # TODO: Cleanup this and db.search_pacakges call sig.
    include_prerelease = request.args.get('includePrerelease', default=False)
    order_by = request.args.get('$orderBy', default='Id')
    filter_ = request.args.get('$filter', default=None)
    top = request.args.get('$top', default=30)
    skip = request.args.get('$skip', default=0)
    sem_ver_level = request.args.get('semVerLevel', default='2.0.0')
    search_query = request.args.get('searchTerm', default=None)
    target_framework = request.args.get('targetFramework', default='')

    # Some of the terms have quotes surrounding them
    search_query = search_query.strip("'")
    target_framework = target_framework.strip("'")

    results = db.search_packages(session,
                                 include_prerelease=include_prerelease,
                                 #order_by=order_by,
                                 filter_=filter_,
                                 search_query=search_query,
                                 )

    feed = FeedWriter('Search', request.url_root)
    resp = make_response(feed.write_to_output(results))
    resp.headers['Content-Type'] = FEED_CONTENT_TYPE_HEADER

    # Keep this line for future debugging. Will fill up the logs *very*
    # quickly if there are a lot of packages on the server.
    # logger.debug(resp.data.decode('utf-8').replace('><',' >\n<'))

    return resp


@pages.route('/updates', methods=['GET'])
def updates():
    """
    I thought this was `nuget restore` but that looks to be
    `GET http://localhost:5000/Packages(Id='xunit',Version='2.3.1')`
    """
    logger.debug("Route: /updates")
    ids = request.args.get('packageids').strip("'").split('|')
    versions = request.args.get('versions').strip("'").split('|')
    include_prerelease = request.args.get('includeprerelease', default=False)
    pkg_to_vers = {k: v for k, v in zip(ids, versions)}

    results = db.package_updates(session, pkg_to_vers, include_prerelease)

    feed = FeedWriter('GetUpdates', request.url_root)
    resp = make_response(feed.write_to_output(results))
    resp.headers['Content-Type'] = FEED_CONTENT_TYPE_HEADER
    return resp



# -*- coding: utf-8 -*-
"""
Logging setup and handlers.
"""
import zlib
import logging
import os
import time
from pathlib import Path
from logging.handlers import RotatingFileHandler


LOG_LEVEL_BASE = logging.DEBUG
LOG_LEVEL_FILE = LOG_LEVEL_BASE
LOG_LEVEL_CONSOLE = logging.DEBUG
LOG_LEVEL_GUI = LOG_LEVEL_BASE

LOG_FMT = ("%(asctime)s.%(msecs)03dZ"
           " [%(levelname)-8.8s]"
           " [%(module)-8.8s]"       # Note implicit string concatenation.
           " [%(funcName)-16.16s]"
           "  %(message)s"
           )
DATE_FMT = "%Y-%m-%dT%H:%M:%S"


class CustomLoggingFormatter(logging.Formatter):
    """
    Custom logging formatter. Overrides funcName and module if a value
    for name_override or module_override exists.
    """
    # Make sure things are saved in UTC.
    converter = time.gmtime

    def format(self, record):
        if hasattr(record, 'name_override'):
            record.funcName = record.name_override
        if hasattr(record, 'module_override'):
            record.module = record.module_override

        return super(CustomLoggingFormatter, self).format(record)


# Custom namer and rotator.
# Taken from https://docs.python.org/3/howto/logging-cookbook.html
def _gzip_namer(name):
    return name + ".gz"


def _gzip_rotator(source, dest):
    # Compress the source file into our dest
    with open(source, 'rb') as open_source:
        data = open_source.read()
        compressed = zlib.compress(data)
        with open(dest, 'wb') as open_dest:
            open_dest.write(compressed)

    # Keep a copy of the most recent rotation uncompressed to make things
    # easier for users.
    os.replace(source, dest + ".1")

    os.remove(source)


def setup_logging(to_console=True, to_file=False, log_path=None):
    # Create the logger
    logger = logging.getLogger('pynuget')
    logger.setLevel(LOG_LEVEL_BASE)

    if to_console:
        _setup_console_logging(logger)

    if to_file:
        _setup_file_logging(logger, log_path)

    return logger


def _setup_file_logging(logger, log_path):
    # Create paths and files as needed.
    log_path = Path(log_path)
    os.makedirs(str(log_path.parent), mode=0x0755, exist_ok=True)
    log_path.touch(mode=0o0664, exist_ok=True)

    handler = RotatingFileHandler(str(log_path), maxBytes=1e7)
    handler.rotator = _gzip_rotator
    handler.namer = _gzip_namer
    handler.setLevel(LOG_LEVEL_FILE)
    formatter = CustomLoggingFormatter(LOG_FMT, DATE_FMT)
    handler.setFormatter(formatter)
    name = "File Handler"
    handler.set_name(name)
    if name not in [h.name for h in logger.handlers]:
        logger.addHandler(handler)
        logger.info("File logging initialized")


def _setup_console_logging(logger):
    """Set up logging to the console."""
    handler = logging.StreamHandler()
    handler.setLevel(LOG_LEVEL_CONSOLE)
    formatter = CustomLoggingFormatter(LOG_FMT, DATE_FMT)
    handler.setFormatter(formatter)
    name = "Console Handler"
    handler.set_name(name)
    if name not in [h.name for h in logger.handlers]:
        logger.addHandler(handler)
        logger.info("Console logging initialized")



# -*- coding: utf-8 -*-
"""
"""
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop

from glob import glob
from os.path import basename
from os.path import splitext


# #########################################################################
# These classes will only work if the user runs:
#    python setup.py develop
#    python setup.py install
class DevelopCommand(develop):
    """
    Modify the `develop` installation workflow.
    """
    def run(self):
        # Pre-install: nothing
        develop.run(self)
        # Post-install: perhaps init the dirs and database?


class InstallCommand(install):
    """
    Modify the starndard installation workflow.
    """
    def run(self):
        # Pre-install: nothing
        install.run(self)
        # Post-install: perhaps init the dirs and database?

# #########################################################################


classifiers = [
    "Development Status :: 1 - Planning",
]

requires = [
    "sqlalchemy>=1.2.7",
    "flask>=0.12.2",
    "lxml>=4.2.1",
    "requests>=2.19.1",
]

entry_points = {
    'console_scripts': [
        'pynuget = pynuget.cli:main',
    ],
}

with open("README.rst", 'r') as openf:
    long_description = openf.read()

setup(
    # General
    name="pynuget",
    version="0.1.0",

    # Descriptions
    description="A Python-based NuGet Server",
    long_description=long_description,
    long_description_content_type='text/x-rst',

    # Owner
    url='https://github.com/dougthor42/pynuget/',
    author='Douglas Thor',
    author_email='doug.thor@gmail.com',

    # Files
    package_dir={'': 'src'},
    packages=find_packages(where='src'),

    # these get copied to sys.prefix/data, which will be the venv's folder.
    data_files=[
        ('data', ['wsgi.py']),
        ('data', ['apache-example.conf']),
    ],

    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],

    classifiers=classifiers,

    # Versions and Requirements
    python_requires=">=3.5",
    install_requires=requires,
    #tests_requires=

    entry_points=entry_points,

    # https://stackoverflow.com/a/36902139/1354930
    cmdclass={
        'develop': DevelopCommand,
        'install': InstallCommand,
    },
)



# -*- coding: utf-8 -*-
"""
"""

import sys
from argparse import ArgumentParser


from pynuget import commands


SERVER_PATH = '/var/www/pynuget'


def main():
    """
    """
    # Parent parser (common args).
    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        '-v', '--verbose',
        action='count',
        help=("Increase the verbosity. Can be supplied multiple times."
              " Max value: -vv"),
        default=0,
    )

    parent_parser.add_argument(
        '-V', "--version",
        action='store_true',
        help="Print the version number and exit.",
    )

    # Common argument
#    server_path_arg = ArgumentParser(add_help=False)
#    server_path_arg.add_argument(
#        "--server-path",
#        help=("The path that the server should live at. Must be an absolute"
#              " path. Defaults to '/var/www/pynuget'"),
#        default="/var/www/pynuget",
#    )

    # Main parser
    parser = ArgumentParser(
        description="Administer the PyNuGet Server.",
        parents=[parent_parser],
    )

    # Subparsers ###########################################################
    subparser = parser.add_subparsers(
        title='subcommands',
        dest="verb",
    )

    # Init
    parser_init = subparser.add_parser(
        'init',
        help=("Create necessary folders and files. Copy Apache configuration"
              " template if one doesn't already exist."),
        parents=[parent_parser],
    )
    parser_init.set_defaults(func=run_init)
    parser_init.add_argument(
        "--apache-config",
        help=("Define a custom name for the Apache configuation file."
              " Defaults to 'pynuget.conf'."),
        default="pynuget.conf",
    )
    parser_init.add_argument(
        "--package-dir",
        help=("The directory that the packages will be saved in and served"
              " from. If this value is not an absolute path, `server_path`"
              " will be used as the root. Defaults to 'packages' in"
              " `server_path`"),
        default="nuget_packages",
    )
    parser_init.add_argument(
        "--db-name",
        help=("The name of the database to use. If `db_backend` == 'sqlite',"
              " then this is the relative or absolute path to the SQLite file"
              " to use. If `db_backend` != 'sqlite', then this is the name of"
              " the Schema to create. Defaults to 'nuget_packages.sqlite' in"
              " `server_path`."),
        default="nuget_packages.sqlite",
    )
    parser_init.add_argument(
        "--db-backend",
        help=("What database backend to use. Defaults to 'sqlite'."),
        default="sqlite",
        choices=["sqlite", "mysql", "postgresql"],
    )

    # Clear
    parser_clear = subparser.add_parser(
        "clear",
        help=("Delete all packages and empty the package database. Basically"
              " resets things to a fresh install but does not change any"
              " settings."),
        parents=[parent_parser],
    )
    parser_clear.set_defaults(func=run_clear)
    parser_clear.add_argument(
        '-y', "--yes",
        help=("Ignore the confirmation prompt and delete everything."),
        action='store_true',
    )

    # Rebuild
    parser_rebuild = subparser.add_parser(
        "rebuild",
        help=("Rebuild the package database based on the files in"
              " the package directory. Useful when manually adding or"
              " removing NuGet package files."),
        parents=[parent_parser],
    )
    parser_rebuild.set_defaults(func=run_rebuild)

    # Push
    parser_push = subparser.add_parser(
        "push",
        help=("Push a package to the server."),
        parents=[parent_parser],
    )
    parser_push.set_defaults(func=run_push)
    parser_push.add_argument(
        "-s", "--source",
        help=("The url of the server to push to."),
        required=True,
    )
    parser_push.add_argument(
        "-k", "--api-key",
        dest='key',
        help=("The NuGet API key of the server."),
        default='',
        required=False,
    )
    parser_push.add_argument(
        "file",
        help=("The file to upload."),
    )

    # Parse the args
    args = parser.parse_args()

    # Show help when no commands or flags are given.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(2)

    # Run the function
    args.func(args)


def run_init(args):
    commands.init(
        server_path=SERVER_PATH,
        package_dir=args.package_dir,
        db_name=args.db_name,
        db_backend=args.db_backend,
        apache_config=args.apache_config,
    )


def run_clear(args):
    commands.clear(server_path=SERVER_PATH)


def run_rebuild(args):
    commands.rebuild()


def run_push(args):
    commands.push(args.file, args.source, args.key)
